from typing import Dict, Any, Optional, List, Type, Union, Set, Callable

from dataspree.inference.core.config import ConfigFieldTypes, camel_to_snake, subclasses

from abc import ABC, abstractmethod


class ConfigOption(ABC):

    def __init__(self, config_entries: Dict[str, Any], child_str: Optional[str] = None) -> None:
        super(ConfigOption, self).__init__()
        self.name: Optional[str] = config_entries.get('name')
        self.display_name: Optional[str] = config_entries.get('display_name', self.name)
        self.description: Optional[str] = config_entries.get('description')
        self.default: Optional[Any] = config_entries.get('default')
        self.og_default: Optional[Any] = config_entries.get('default')
        self.child_str: Optional[str] = child_str
        self.children: List[ConfigOption] = [parse_config_field(f) for f in config_entries.get(child_str, [])]
        self.warning: Optional[List[Union[str, Exception]]] = None

    @abstractmethod
    def apply(self, config_build: Optional[Union[Dict[str, Any], List, Any]]) -> None:
        """
        Apply the value from the node config as default value
        """

    @abstractmethod
    def set_warning(self, config_build: Optional[Union[Dict[str, Any], List, Any]]) -> None:
        """
        Set problems
        """

    def generate_config_entries(self):
        # Generates the config_entries configuration
        content = {'name': self.name, 'default': self.default}
        if self.warning is not None:
            content['__warning'] = None if self.warning is None else [str(warn) for warn in self.warning]
        if self.child_str:
            children = [child.generate_config_entries() for child in self.children]
            if len(children):
                content[self.child_str] = children
        return content

    def active(self, config_build: Any) -> bool:
        return config_build is not None


class MultipleInput(ConfigOption):  # noqa
    def __init__(self, child_str: str, config: Dict[str, Any]):
        super(MultipleInput, self).__init__(config, child_str=child_str)


class Group(MultipleInput):
    def __init__(self, config: Dict[str, Any]):
        super(Group, self).__init__('fields', config)

    def apply(self, config_build: Optional[Dict[str, Any]]):
        if config_build is not None:
            for child in self.children:
                child.apply(config_build.get(child.name))

    def set_warning(self, warning: Optional[Dict[str, Any]]):
        self.warning = None

        if warning is not None and isinstance(warning, dict):
            self.warning = warning.get('__warning')

            for child in self.children:
                child.set_warning(warning.get(child.name))


class GroupSelection(MultipleInput):
    def __init__(self, config: Dict[str, Any]):
        super(GroupSelection, self).__init__('groups', config)

    def apply(self, config_build: Optional[Dict[str, Any]]) -> None:
        if config_build is None:
            return

        self.default = config_build.get('selection')

        # Set the value for the currently active child.
        # Set the own value (selection)
        for child in self.children:
            if child.name == self.default:
                child.apply(config_build.get(child.name))

    def set_warning(self, warning: Optional[Dict[str, Any]]) -> None:
        self.warning = None

        if warning is None or not isinstance(warning, dict):
            return

        self.warning = warning.get('__warning')

        for child in self.children:
            if child.name == self.default:
                child.set_warning(warning.get(child.name))

    def generate_config_entries(self):
        content = {'name': self.name, 'selection': self.default}
        if self.warning is not None:
            content['__warning'] = None if self.warning is None else [str(warn) for warn in self.warning]
        if self.child_str:
            children = [child.generate_config_entries() if child.name == self.default else {} for child in self.children]
            if len(children):
                content[self.child_str] = children
        return content


class GroupSwitchable(MultipleInput):
    def __init__(self, config: Dict[str, Any]):
        super(GroupSwitchable, self).__init__('fields', config)

    def apply(self, config_build: Optional[Dict[str, Any]]) -> None:
        self.default = config_build is not None and all((config_build.get(child.name) for child in self.children))

        if self.default:
            for child in self.children:
                child.apply(config_build.get(child.name))

    def set_warning(self, warning: Optional[Dict[str, Any]]) -> None:
        self.warning = None

        if self.default and warning is not None and isinstance(warning, dict):
            self.warning = warning.get('__warning')
            for child in self.children:
                child.set_warning(warning.get(child.name))

    def active(self, _config_build: Any) -> bool:
        # can always be active, independent of the argument, because it is active even if the switch is off.
        return True

    def generate_config_entries(self):
        content = {'name': self.name, 'default': self.default}
        if self.warning is not None:
            content['__warning'] = None if self.warning is None else [str(warn) for warn in self.warning]
        if self.child_str and self.default:
            children = [child.generate_config_entries() for child in self.children]
            if len(children):
                content[self.child_str] = children
        return content


class ListInput(MultipleInput):
    def __init__(self, config: Dict[str, Any]):
        super(ListInput, self).__init__('fields', config)

    def apply(self, config_build: Optional[Dict[str, Any]]) -> None:
        self.default = []

        if config_build is not None:
            for cfg in config_build:
                self.default.append([])
                for child in self.children:
                    child.apply(cfg.get(child.name))
                    self.default[-1].append(child.generate_config_entries())

    def set_warning(self, warning: Optional[Dict[str, Any]]) -> None:
        self.warning = None

        if warning is not None and isinstance(warning, dict):
            self.warning = warning.get('__warning')

            for cfg in warning:
                for child in self.children:
                    child.set_warning(cfg.get(child.name))

    def generate_config_entries(self):
        content = {'name': self.name, 'default': self.default}
        if self.warning is not None:
            content['__warning'] = None if self.warning is None else [str(warn) for warn in self.warning]

        return content


class SelectionDropdown(MultipleInput):
    def __init__(self, config: Dict[str, Any]):
        super(SelectionDropdown, self).__init__('fields', config)

    def apply(self, config_build: Any):
        self.default = config_build

    def set_warning(self, warning: Any):
        self.warning = None if warning is None or not isinstance(warning, dict) else warning.get('__warning')


class ScalarInput(ConfigOption):

    def __init__(self, config: Dict[str, Any]):
        super(ScalarInput, self).__init__(config)

    def apply(self, config_build: Any):
        self.default = config_build

    def set_warning(self, warning: Any):
        self.warning = None if warning is None or not isinstance(warning, dict) else warning.get('__warning')


class BoolInput(ScalarInput):
    ...


class DateInput(ScalarInput):
    ...


class DoubleInput(ScalarInput):
    ...


class IntegerInput(ScalarInput):
    ...


class JsonTextInput(ScalarInput):
    ...


class StringInput(ScalarInput):
    ...


class PasswordInput(ScalarInput):
    ...


class TextAreaInput(ScalarInput):
    ...


__config_field_values: Set[str] = {v.value for v in ConfigFieldTypes}  # noqa
__config_field_lookup: Dict[str, Type[ConfigOption]] = {
    camel_to_snake(subclass.__name__): subclass for subclass in subclasses(ConfigOption)
    if camel_to_snake(subclass.__name__) in __config_field_values}
assert len(__config_field_lookup) == len(__config_field_values), 'ConfigFieldValue not implemented.'


def parse_config_field(config: Dict[str, Any]) -> ConfigOption:
    config_type: ConfigFieldTypes = config.get('type', ConfigFieldTypes.GROUP)
    config_type_str: str = config_type.value
    return __config_field_lookup[config_type_str](config)
