import logging
import time
import numba

import numpy as np
from scipy import ndimage as sndimage

from typing import List, Optional, Tuple, Union, Any, Dict, Callable, Set
from numba.np.extensions import cross2d

from dataspree.inference.core.environment import is_application_debug
from dataspree.inference.core.item import Item
from dataspree.inference.core.producer import Property, PropertyType
from dataspree.inference.core.simple_plugin import SimplePlugin

logger = logging.getLogger(__name__)


class Localization(SimplePlugin):

    def __init__(self, name: str, xyz_matrix_path: Optional[str] = None, default_object_depth: float = 1.0,
                 confidence_threshold: float = 0.7) -> None:
        super().__init__(name=name, properties=[
            Property('point_cloud', PropertyType.POINT_CLOUD),
            Property('image', PropertyType.IMAGE),
            Property('inference', PropertyType.INFERENCE)])

        self.default_object_depth: float = default_object_depth
        self.xyz_matrix_path: Optional[str] = xyz_matrix_path
        self.xyz_matrix: Optional[np.ndarray] = None if self.xyz_matrix_path is None else \
            np.fromfile(self.xyz_matrix_path, dtype=np.float32, offset=3 * 4).reshape(
                np.fromfile(self.xyz_matrix_path, dtype=np.int32, count=3))

        self.confidence_threshold: float = confidence_threshold

        self.__tmp_box_points: Optional[np.ndarray] = None
        self.__tmp_box_hull_points: Optional[np.ndarray] = None

    def process_item(self, item: Item) -> Item:
        raise NotImplementedError()
        return item
        """
        t_s = time.time()
        detections: Optional[List] = item.get(['inference', 'detection'])
        if detections is None:
            logger.warning('No detection inference worker seems to be preceeding the Localization plugin. '
                           'Localization cannot be performed.')
            return item

        xyz_matrix: Optional[np.ndarray] = item['point_cloud'] if self.xyz_matrix is None else self.xyz_matrix
        if xyz_matrix is None:
            logger.warning('No reference point cloud specified.'
                           'Localization cannot be performed ')
            return item

        confidence_image: Optional[np.ndarray] = item.get('confidence_image')

        image_height, image_width = xyz_matrix.shape[:2]
        assert xyz_matrix.shape == (image_height, image_width, 3)

        # Get upper triangle (bottom left, top left, top right) and unrotated enclosing rectangle (min max).
        vectors: np.ndarray
        to_original_index: np.ndarray
        vectors, to_original_index = dataspree_inference_boxes_to_vectors(
            detections, image_width=image_width, image_height=image_height, threshold=self.confidence_threshold)

        # print(f'extracted {len(vectors)} detections')
        number_of_detections: int = len(vectors)
        assert vectors.shape == (number_of_detections, 5, 2), vectors.shape
        assert to_original_index.shape == (number_of_detections,)

        if self.__tmp_box_points is None or self.__tmp_box_points.shape != (
                xyz_matrix.shape[0] * xyz_matrix.shape[1], *xyz_matrix.shape[2:]):
            self.__tmp_box_points = np.empty((xyz_matrix.shape[0] * xyz_matrix.shape[1], *xyz_matrix.shape[2:]),
                                             dtype=np.float32)
            self.__tmp_box_hull_points = np.empty((xyz_matrix.shape[0] * xyz_matrix.shape[1], *xyz_matrix.shape[2:]),
                                                  dtype=np.float32)

        for original_box_index, box_representation in zip(to_original_index, vectors):
            self.__tmp_box_points, self.__tmp_box_hull_points = \
                self.get_detection_mask(xyz_image=xyz_image, box_representation=box_representation,
                                        box_points=self.__tmp_box_points, box_border_points=self.__tmp_box_hull_points,
                                        detection=detections[original_box_index], binary_image=None,
                                        confidence_image=confidence_image)


        # print('finish process', item['inference']['detection'])
        print(f'{time.time() - t_s:0.3f}took ')
        return item
        """

    @staticmethod
    def get_detection_masks(detections: List[Dict], xyz_image: np.ndarray, confidence_image: np.ndarray,

                            labels_accumulate: Optional[Set[Any]] = None,

                            remove_by_cluster: bool = False,

                            confidence_threshold: float = 0.0,
                            confidence_threshold_label: Optional[Dict[Any, float]] = None,

                            z_threshold: Optional[float] = 20,
                            z_threshold_label: Optional[Dict[Any, float]] = None,

                            plot_identifier: Optional[int] = None, plot_show: bool = False,

                            box_points: Optional[np.ndarray] = None,
                            box_border_points: Optional[np.ndarray] = None) -> Tuple[np.ndarray, np.ndarray]:
        """
        Extract the detection masks in the image.

        :param detections:         List of inference detection dictionaries.
        :param xyz_image:          Matrix that contains the xyz image. Shape: [image_height, image_width, 3].
        :param confidence_image:   Matrix that contains the confidence image. Shape: [image_height, image_width].

        :param labels_accumulate:  Set of all classes that are to be treated as segmentation detections (to be used
                                   when the detection quality is lacking because it is hard to discern where
                                   the boundaries of detections are).

        :param remove_by_cluster:  Set true, to cluster the 3D points and utilize only the points of the largest
                                   cluster.


        :param confidence_threshold:        Default confidence threshold for filtering out detections
        :param confidence_threshold_label:  Confidence threshold for instances of specific classes, preferred over the
                                            default confidence threshold.

        :param z_threshold:        Filter out points deviating from median z position by threshold if set and > 0.
        :param z_threshold_label:  Z Threshold for instances of specific classes, preferred over the default threshold.

        :param plot_identifier:   Set to a value != None for exporting debug visualizations to disk for debug purposes.
        :param plot_show:         Set to True to open matplotlib window with debug visualizations during processing.

        :param box_points:          Temp array for not needing reallocation.
        :param box_border_points:   Temp array for not needing reallocation.
        """
        image_height: int = confidence_image.shape[0]
        image_width: int = confidence_image.shape[1]
        number_of_pixels: int = image_height * image_width
        assert xyz_image.shape == (image_height, image_width, 3)
        assert confidence_image.shape == (image_height, image_width)

        # Initialize box point temporary array if need-be
        box_points, box_border_points = initialize_box_point_array(box_points, box_border_points, number_of_pixels)

        # Get upper triangle (bottom left, top left, top right) and minimum unrotated enclosing rectangle (min max).
        times = [(time.perf_counter(), '')]
        dataspree_inference_boxes_to_vectors(detections, image_width=image_width, image_height=image_height,
                                             threshold=confidence_threshold,
                                             specific_threshold=confidence_threshold_label)
        times.append((time.perf_counter(), 'inference_boxes_to_vectors'))

        # Create box_occupation images for detections that are used as segmentations
        box_occupation_images: Dict[Any, np.ndarray] = dict()
        for label in labels_accumulate:
            box_occupation_images[label] = np.zeros(confidence_image.shape, dtype=np.uint64)

        #
        # Get detection masks and accumulate detections for those classes that are to be accumulated.
        for detection_index, detection in enumerate(detections):
            label = detection['label']

            if labels_accumulate is not None and label in labels_accumulate:
                Localization.get_detection_mask_accumulation(box_occupation_image=box_occupation_images[label],
                                                             detection=detection)
                times.append((time.perf_counter(), 'accumulate'))

            else:
                current_z_threshold: Optional[float] = z_threshold if z_threshold_label is None else \
                    z_threshold_label.get(label, z_threshold)

                current_plot_identifier: Optional[str] = f'{plot_identifier}.{detection["id"]}' \
                    if plot_identifier is not None else None

                box_points, box_border_points = Localization.get_detection_mask(
                    detection=detection, xyz_image=xyz_image, confidence_image=confidence_image,
                    box_points=box_points, box_border_points=box_border_points,

                    remove_by_cluster=remove_by_cluster, z_threshold=current_z_threshold,

                    plot_identifier=current_plot_identifier, plot_show=plot_show)
                times.append((time.perf_counter(), 'detection_mask'))

        #
        # Get Detection Masks for accumulated detections
        # box_occupation_image = (box_occupation_image * difference_image) > 10 XXX:
        for label, box_occupation_image in box_occupation_images.items():

            current_z_threshold: Optional[float] = z_threshold if z_threshold_label is None else \
                z_threshold_label.get(label, z_threshold)

            if box_occupation_image.any():
                it = 5
                box_occupation_image = sndimage.binary_dilation(box_occupation_image, iterations=it * 2)
                box_occupation_image = sndimage.binary_erosion(box_occupation_image, iterations=it)

                mat_labeled, nr_objects = sndimage.label(box_occupation_image)

                # assign the points
                for identifier in range(nr_objects):
                    current_binary_image = mat_labeled == identifier + 1

                    objs = sndimage.find_objects(current_binary_image)
                    y_start, y_stop = objs[0][0].start, objs[0][0].stop
                    x_start, x_stop = objs[0][1].start, objs[0][1].stop
                    width = x_stop - x_start
                    height = y_stop - y_start
                    x = (x_start + x_stop) / 2
                    y = (y_start + y_stop) / 2

                    # Create box representation for this unrotated box.
                    bottom_left = np.array([x_start, y_stop], dtype=float)
                    top_left = np.array([x_start, y_start], dtype=float)
                    top_right = np.array([x_stop, y_start], dtype=float)
                    min_values = np.array([x_start, y_start], dtype=int)
                    max_values = np.array([x_stop, y_stop], dtype=int)
                    box_representation = np.stack((bottom_left, top_left, top_right, min_values, max_values), axis=0)
                    np.clip(box_representation, [0, 0], [image_width - 1, image_height - 1], out=box_representation)

                    enclosing_rectangle = np.round(box_representation[3:]).astype(int)
                    min_x, min_y = enclosing_rectangle[0, :2]
                    max_x, max_y = enclosing_rectangle[1, :2]
                    c_x, c_y = int(round((min_x + max_x) / 2.0)), int(round((min_y + max_y) / 2.0))
                    assert 0 <= min_x <= c_x <= max_x <= image_width, (0, min_x, max_x, image_width)
                    assert 0 <= min_y <= c_y <= max_y <= image_height, (0, min_y, max_y, image_height)

                    lookup_image = np.zeros(current_binary_image.shape, dtype=np.uint8)[y_start:y_stop, x_start:x_stop]

                    # pose3d, size3d = result #spaeter
                    new_detection_identifier: int = len(detections)

                    detection = {
                        'id': new_detection_identifier,
                        'timestamp': detections[0]['timestamp'],
                        'x': x / image_width,
                        'y': y / image_height,
                        'width': width / image_width,
                        'height': height / image_height,
                        'orientation': 0,
                        'class_confidences': dict(),
                        'label': label,
                        'confidence': 1,
                        'localization': {
                            'rotated_box': box_representation[:3],
                            'enclosing_rectangle': enclosing_rectangle,
                            'min_x': min_x,
                            'min_y': min_y,
                            'max_x': max_x,
                            'max_y': max_y,
                            'c_x': c_x,
                            'c_y': c_y,
                            'masked': True
                        }
                    }

                    detections.append(detection)

                    current_plot_identifier: Optional[str] = f'{plot_identifier}.sup.{detection["id"]}' \
                        if plot_identifier is not None else None

                    box_points, box_border_points = Localization.get_detection_mask(
                        detection=detection, xyz_image=xyz_image, confidence_image=confidence_image,
                        box_points=box_points, box_border_points=box_border_points,

                        remove_by_cluster=remove_by_cluster, z_threshold=current_z_threshold,

                        plot_identifier=current_plot_identifier, plot_show=plot_show,

                        lookup_image=lookup_image)

                    # plot_show = True

                    if plot_show or current_plot_identifier:
                        import matplotlib.pyplot as plt
                        current_xyz_image = xyz_image[y_start:y_stop, x_start:x_stop]
                        plt.figure(figsize=(12 / 2, 12 / 2))
                        plt.subplots_adjust(left=0, bottom=0, right=1, top=0.95, wspace=0.01, hspace=0.01)

                        plt.subplot(2, 2, 1)
                        plt.imshow(lookup_image)

                        plt.subplot(2, 2, 2)
                        plt.imshow(((lookup_image == 2) * current_xyz_image[..., -1]))

                        plt.subplot(2, 2, 3)
                        plt.imshow(((lookup_image >= 1) * current_xyz_image[..., -1]))

                        plt.subplot(2, 2, 4)
                        plt.scatter(*np.where((lookup_image >= 1)), s=0.5, color='blue')
                        plt.scatter(*np.where((lookup_image >= 2)), s=0.5, color='red')
                        plt.scatter(*np.where((lookup_image == 0)), s=0.5, color='black')

                        if plot_show:
                            plt.show()

                        if current_plot_identifier is not None:
                            plt.savefig(f'{plot_identifier}_get_detection_mask_assemble_lookup_image.png')

                    times.append((time.perf_counter(), 'det_from_seg'))

        logger.debug('Localization: get detection masks'
                    + ', '.join([f'{v2}: {int((t2 - t1) * 1000) / 1000}'
                                 for (t1, _), (t2, v2) in zip(times[:-1], times[1:])])
                    + f' total: {times[-1][0] - times[0][0]}')

        return box_points, box_border_points

    @staticmethod
    def get_detection_mask_accumulation(box_occupation_image: np.ndarray, detection: Dict) -> bool:
        """
        For detections that are not accurate in position, use the detection zoo to create a heatmap to later discern
        extract detections from the detection heatmap.

        :param box_occupation_image: Image that contains the weighted count of the boxes spanning this area.
        :param detection:            Detection dictionary that contains one detection.
        """

        image_height: int = box_occupation_image.shape[0]
        image_width: int = box_occupation_image.shape[1]
        assert box_occupation_image.shape == (image_height, image_width)

        object_mask: Optional[np.ndarray] = get_object_mask(detection=detection,
                                                            image_width=image_width, image_height=image_height)
        if object_mask is None:
            return False

        localization: Dict = detection['localization']
        min_x, min_y = localization['min_x'], localization['min_y']
        max_x, max_y = localization['max_x'], localization['max_y']

        box_occupation_image_section = box_occupation_image[min_y:max_y, min_x:max_x]
        assert box_occupation_image_section.shape == object_mask.shape, \
            (box_occupation_image_section.shape, object_mask.shape)

        box_occupation_image_section += (object_mask.astype(np.uint64) * int(1 + detection['confidence'] / 0.2))
        return True

    @staticmethod
    def get_detection_mask(detection: Dict,
                           xyz_image: np.ndarray, confidence_image: Optional[np.ndarray],
                           *,
                           remove_by_cluster: bool = True,
                           z_threshold: Optional[float] = 20,

                           plot_identifier: Optional[str] = None, plot_show: bool = False,

                           box_points: Optional[np.ndarray] = None,
                           box_border_points: Optional[np.ndarray] = None,

                           lookup_image: Optional[np.ndarray] = None):
        """
        Extract the detection masks in the image.

        :param detection:          Inference detection dictionary.

        :param xyz_image:         Matrix that contains the xyz image. Shape: [image_height, image_width, 3].
        :param confidence_image:   Matrix that contains the confidence image. Shape: [image_height, image_width].

        :param remove_by_cluster:       Set true, to cluster the 3D points and utilize only the points of the largest
                                        cluster.

        :param z_threshold:             Filter out points deviating from median z position by threshold if set and > 0.


        :param plot_identifier:   Set to a value != None for exporting debug visualizations to disk for debug purposes.
        :param plot_show:         Set to True to open matplotlib window with debug visualizations during processing.

        :param box_points:          Temp array for not needing reallocation.
        :param box_border_points:   Temp array for not needing reallocation.

        :param lookup_image:        Image used for visualization or None.
        """

        #
        # Threshold: make sure that everything is in the expected shape
        image_height: int = xyz_image.shape[0]
        image_width: int = xyz_image.shape[1]
        number_of_pixels: int = image_height * image_width
        assert xyz_image.shape == (image_height, image_width, 3), xyz_image.shape

        if confidence_image is None:
            confidence_image = np.ones((image_height, image_width), dtype=np.uint16)
        assert confidence_image.shape == (image_height, image_width)

        # Initialize box point temporary array if need-be
        box_points, box_border_points = initialize_box_point_array(box_points, box_border_points, number_of_pixels)

        localization: Optional[Dict] = detection.get('localization')
        if localization is None:
            return box_points, box_border_points

        min_x, min_y = localization['min_x'], localization['min_y']
        max_x, max_y = localization['max_x'], localization['max_y']
        assert 0 <= min_x <= max_x < image_width
        assert 0 <= min_y <= max_y < image_height

        #
        # Get the binary object mask
        object_mask = get_object_mask(detection, image_width=image_width, image_height=image_height)
        assert object_mask is not None
        assert object_mask.shape == (max_y - min_y, max_x - min_x), object_mask.shape

        localization_target = localization['3d'] = dict()

        # Compute the median distance within the object_mask excluding invalid pixels (distance == 0).
        # Therefore the condition is not wrt. object_mask, but wrt masked_distance
        masked_distance = xyz_image[min_y:max_y, min_x:max_x, -1] * object_mask
        localization_target['median_z'] = median_z = 1.0 * np.median(masked_distance[masked_distance != 0])

        # 1) Extract points
        #   a) all the points that belong to the box
        #   a) all the points that belong to the box' border
        number_of_box_points: int
        number_of_box_border_points: int

        number_of_box_points, number_of_box_border_points, cluster = extract_box_pixels(
            object_mask=object_mask, xyz_image=xyz_image, confidence_image=confidence_image,
            box_points=box_points, box_border_points=box_border_points,
            z_threshold=float(z_threshold if z_threshold is not None else 0.0), median_z=median_z,
            remove_by_cluster=remove_by_cluster, min_x=min_x, min_y=min_y, lookup_image=lookup_image)

        assert not np.any(np.isnan(box_points[:number_of_box_points]))
        assert not np.any(np.isnan(box_border_points[:number_of_box_border_points]))

        localization_target['cluster'] = cluster
        localization_target['box_points'] = box_points[:number_of_box_points].copy()
        localization_target['box_border_points'] = box_border_points[:number_of_box_border_points].copy()

        # Numpy array, in which elements are true (i.e., >= 1) if they might be border points if they are in range.
        if plot_identifier is not None or plot_show:
            import matplotlib.pyplot as plt

            plt.figure(figsize=(3, 3))
            plt.subplots_adjust(left=0, bottom=0, right=1, top=0.95, wspace=0.01, hspace=0.01)

            plt.subplot(4, 2, 1)
            plt.imshow(object_mask)

            plt.subplot(4, 2, 2)
            plt.plot(*np.array([[min_x, min_y], [min_x, max_y], [max_x, max_y],
                                [max_x, min_y], [min_x, min_y]]).T)
            plt.imshow(xyz_image)

            ax = plt.subplot(2, 1, 2)
            ax.scatter(*box_border_points[:number_of_box_border_points, :2].T, color='blue', s=50, alpha=0.5)
            ax.scatter(*box_points[:number_of_box_points, :2].T, color='green', s=5)
            ax.set_aspect('equal')

            if plot_show:
                plt.show()
            else:
                plt.savefig(f'{plot_identifier}_get_detection_mask.png')

        return box_points, box_border_points

    @staticmethod
    def filter_detection_masks(detections: List[Dict], binary_image: Optional[np.ndarray] = None,
                               min_ratio_change: Optional[float] = None, use_cached: bool = False) -> List[Dict]:
        """
        :param detections:              List of inference detection dictionaries.
        :param binary_image:            Mask image or None if no mask is to be applied.
        :param min_ratio_change:  Filter out entire box, if ratio_change_threshold is not None and the percentage
                                        of box-pixels inside the #binary_image is smaller than this threshold.
        :param use_cached:              If True and the ratio change threshold has already been computed for a
                                        detection, use the already computed value. Does not check irf the binary
                                        image changed.
        """

        def threshold_met(ratio_change):
            return min_ratio_change is None or \
                   (ratio_change is not None and ratio_change >= min_ratio_change)

        return [detection for detection in detections if threshold_met(get_ratio_change(detection=detection,
                                                                                        binary_image=binary_image,
                                                                                        use_cached=use_cached))]

    @staticmethod
    def localize_detections_3d(detections: List[Dict], object_depth_comparator_pc: Optional[np.ndarray],
                               default_object_depth: float = 0.0, plot_identifier: Optional[int] = None,
                               plot_show: bool = False, use_cached: bool = True):
        """
        Get size and location in 3d of detections

        :param detections:              List of inference detection dictionaries.
        :param default_object_depth:    Object depth to be written into the size array.
        :param plot_identifier:   Set to a value != None for exporting debug visualizations to disk for debug purposes.
        :param plot_show:         Set to True to open matplotlib window with debug visualizations during processing.
        """

        assert object_depth_comparator_pc is None or (len(object_depth_comparator_pc.shape) == 3
                                                      and object_depth_comparator_pc.shape[-1] == 3)
        # Will contain Id, width, height
        extracted_detections = np.empty((len(detections), 6), dtype=np.float32)
        n_extracted_detections: int = 0
        for detection_index, detection in enumerate(detections):
            current_plot_identifier: Optional[str] = f'{plot_identifier}.sup.{detection["id"]}' \
                if plot_identifier is not None else None

            extract_3d(detection=detection, default_object_depth=default_object_depth,
                       plot_identifier=current_plot_identifier, plot_show=plot_show, use_cached=use_cached,
                       object_depth_comparator_pc=object_depth_comparator_pc)

            if detection.get('size3d') is None:
                logger.debug('size3d is None. Skipped')
                continue

            if detection.get('localization') is None:
                logger.debug('localization is None. Skipped')
                continue

            localization = detection.get('localization')

            size = detection['size3d']
            extracted_detections[n_extracted_detections] = (detection_index, *size[:3], localization.get('c_x', -1),
                                                            localization.get('c_y', -1))
            n_extracted_detections += 1

        extracted_detections = extracted_detections[:n_extracted_detections]
        assert extracted_detections.shape == (n_extracted_detections, 6)
        return extracted_detections


def initialize_box_point_array(box_points: Optional[np.ndarray],
                               box_border_points: Optional[np.ndarray], number_of_pixels: int) \
        -> Tuple[np.ndarray, np.ndarray]:
    """
    Helper function to initialize the numpy helper arrays box_points and box_border_points if necessary.

    """
    # If the temporary matrices need to be updated or initialized, reinstantiate them
    if box_points is None or box_points.shape[0] <= number_of_pixels:
        box_points = np.empty((number_of_pixels, 4), dtype=np.float32)
        box_border_points = np.empty((number_of_pixels, 4), dtype=np.float32)

    assert box_points.shape == (number_of_pixels, 4), (box_points.shape, box_border_points.shape)
    assert box_border_points.shape == (number_of_pixels, 4), (box_points.shape, box_border_points.shape)
    return box_points, box_border_points


def get_ratio_change(detection: Dict, binary_image: Optional[np.ndarray] = None,
                     use_cached: bool = True) -> Optional[float]:
    """
    :param detection:               Inference detection dictionary.
    :param binary_image:            Mask image or None if no mask is to be applied.
    :param use_cached:              If True and the ratio change threshold has already been computed for a
                                    detection, use the already computed value. Does not check irf the binary
                                    image changed.
    """

    localization: Optional[Dict] = detection.get('localization')
    if localization is None:
        return None

    if use_cached and 'ratio_change' in localization:
        return localization['ratio_change']

    image_height: int = binary_image.shape[0]
    image_width: int = binary_image.shape[1]

    min_x, min_y = localization['min_x'], localization['min_y']
    max_x, max_y = localization['max_x'], localization['max_y']
    assert 0 <= min_x <= max_x < image_width
    assert 0 <= min_y <= max_y < image_height

    #
    # Get the binary object mask
    object_mask = get_object_mask(detection, image_width=image_width, image_height=image_height)
    assert object_mask is not None
    assert object_mask.shape == (max_y - min_y, max_x - min_x), (object_mask.shape, max_y - min_y, max_x - min_x)
    ratio_change: float
    if binary_image is None:
        ratio_change = 1.0
    else:
        binary_image_section = binary_image[min_y:max_y, min_x:max_x]
        sum_object_mask = np.sum(object_mask)
        if sum_object_mask == 0:
            ratio_change = 0
        else:
            ratio_change = 1.0 * np.sum(object_mask[binary_image_section > 0]) / sum_object_mask
    assert 0.0 <= ratio_change <= 1.0, ratio_change

    # ratio_change: float = np.mean(encoded_grid_in_range)
    localization['ratio_change'] = ratio_change
    return ratio_change


def get_object_mask(detection: Dict, *, image_width: int, image_height: int, use_cached: bool = True) \
        -> Optional[np.ndarray]:
    """
    Compute the object mask in the image, that incorporates rotations.
    """

    localization: Optional[Dict] = detection.get('localization')
    if localization is None:
        return None

    if localization.get('object_mask') is not None and use_cached:
        min_x, min_y = localization['min_x'], localization['min_y']
        max_x, max_y = localization['max_x'], localization['max_y']

        assert localization.get('object_mask').shape == (max_y - min_y, max_x - min_x)
        return localization.get('object_mask')

    upper_left_triangle: Optional[np.ndarray] = localization.get('rotated_box')
    enclosing_rectangle: Optional[np.ndarray] = localization.get('enclosing_rectangle')
    if upper_left_triangle is None or enclosing_rectangle is None:
        return None

    # The upper left triangle is (bottom left, top left, top right)
    assert upper_left_triangle.shape == (3, 2), upper_left_triangle.shape

    # Select only the relevant part of the point cloud in which parts of the rotated box are situated
    # (min_values, max_values).
    assert enclosing_rectangle.shape == (2, 2)

    min_x, min_y = localization['min_x'], localization['min_y']
    max_x, max_y = localization['max_x'], localization['max_y']
    assert 0 <= min_x <= max_x < image_width
    assert 0 <= min_y <= max_y < image_height, (0, min_y, max_y, image_height, image_width)

    # Construct a grid, in which only the relevant locations in the point cloud are located
    grid: np.ndarray = np.mgrid[min_y:max_y, min_x:max_x]
    assert grid.shape == (2, max_y - min_y, max_x - min_x), grid.shape

    # A flat view of the grid, that we encode
    grid_view_flat = grid.view()
    grid_view_flat.shape = (2, -1)
    grid_view_flat = grid_view_flat.T

    # Get the functions that transform from standard coordinate system to the coordinate system spanned by the
    # upper triangle cathetus-vectors.
    encode, _decode = coordinate_transformation(upper_left_triangle[..., ::-1])

    # Rotate the grid to discern the points inside the rotated box from those outside
    encoded_grid = encode(grid_view_flat)
    assert encoded_grid.shape == grid_view_flat.shape, (encoded_grid.shape, grid_view_flat.shape)

    encoded_grid = encoded_grid.T.reshape(2, max_y - min_y, max_x - min_x)
    assert encoded_grid.shape == (2, max_y - min_y, max_x - min_x)

    encoded_grid_in_range = np.prod((encoded_grid >= 0) * (encoded_grid <= 1), axis=0)
    assert encoded_grid_in_range.shape == (max_y - min_y, max_x - min_x)

    localization['object_mask'] = encoded_grid_in_range
    return encoded_grid_in_range


def get_area_higher_lower(detection: Dict, *, distance_image_previous: np.ndarray,
                          distance_image: np.ndarray) -> Optional[Tuple[np.ndarray, np.ndarray, np.ndarray]]:
    localization: Optional[Dict] = detection.get('localization')
    if localization is None:
        return None

    min_x, min_y = localization['min_x'], localization['min_y']
    max_x, max_y = localization['max_x'], localization['max_y']

    object_mask: np.ndarray = localization['object_mask']
    if object_mask is None:
        return None

    assert 0 <= min_x <= max_x < distance_image_previous.shape[1]
    assert 0 <= min_y <= max_y < distance_image_previous.shape[0]
    assert object_mask.shape == (max_y - min_y, max_x - min_x)
    assert distance_image_previous.shape == distance_image.shape
    assert len(distance_image.shape) == 2

    relevant_distance_image_previous = distance_image_previous[min_y: max_y, min_x: max_x] * object_mask
    relevant_distance_image = distance_image[min_y: max_y, min_x: max_x] * object_mask
    assert relevant_distance_image_previous.shape == relevant_distance_image.shape
    assert len(relevant_distance_image.shape) == 2

    show = False  # XXX
    if show:
        import matplotlib.pyplot as plt
        plt.subplot(221)
        plt.imshow(relevant_distance_image_previous)
        plt.subplot(222)
        plt.imshow(relevant_distance_image)
        plt.show()

    valid: np.ndarray = (relevant_distance_image != 0) * (relevant_distance_image_previous != 0)
    assert valid.shape == relevant_distance_image.shape

    relevant_distance_image_previous = relevant_distance_image_previous[valid]
    relevant_distance_image = relevant_distance_image[valid]
    num_relevant_pixel: int = relevant_distance_image.shape[0]
    assert len(relevant_distance_image.shape) == 1
    assert relevant_distance_image_previous.shape == (num_relevant_pixel,)
    assert relevant_distance_image.shape == (num_relevant_pixel,)

    significant_distance: int = 50  # 5cm is considered a significant change.

    # area: area significantly lower (distance larger) than previously
    # distance now significantly larger than previously
    px_lower_than_previously = relevant_distance_image_previous + significant_distance < relevant_distance_image

    # area significantly higher (distance smaller) than previously
    # distance now significantly smaller than previously
    px_higher_than_previously = relevant_distance_image_previous > relevant_distance_image + significant_distance

    return relevant_distance_image, px_lower_than_previously, px_higher_than_previously


def is_old_detection_now_occluded(detection: Dict, *, distance_image_previous: np.ndarray, distance_image: np.ndarray) \
        -> Optional[bool]:
    r = get_area_higher_lower(detection, distance_image_previous=distance_image_previous, distance_image=distance_image)

    if r is None:
        return None

    relevant_distance_image, px_lower_than_previously, px_higher_than_previously = r
    num_relevant_pixel: int = relevant_distance_image.shape[0]

    return (np.sum(px_higher_than_previously) > max(0.05 * num_relevant_pixel, 20)) \
           and not (np.sum(px_lower_than_previously) > max(0.05 * num_relevant_pixel, 20))


def was_new_detection_previously_occluded(detection: Dict, *, distance_image_previous: np.ndarray,
                                          distance_image: np.ndarray) -> Optional[bool]:
    r = get_area_higher_lower(detection, distance_image_previous=distance_image_previous, distance_image=distance_image)

    if r is None:
        return None

    relevant_distance_image, px_lower_than_previously, px_higher_than_previously = r
    num_relevant_pixel: int = relevant_distance_image.shape[0]

    return np.sum(px_lower_than_previously) > max(0.05 * num_relevant_pixel, 20)


@numba.jit(numba.types.Tuple((numba.int32, numba.int32, numba.uint16[:, :]))(
    numba.int64[:, :], numba.int16[:, :, :], numba.types.Array(numba.types.uint16, 2, 'A', readonly=True),
    numba.float32[:, :], numba.float32[:, :],
    numba.float64, numba.float64, numba.boolean, numba.int64, numba.int64,
    numba.optional(numba.uint8[:, :])), nopython=True, parallel=False)
def extract_box_pixels(object_mask: np.ndarray, xyz_image: np.ndarray, confidence_image: np.ndarray,
                       box_points: np.ndarray, box_border_points: np.ndarray,
                       z_threshold: float, median_z: float, remove_by_cluster: bool, min_x: int, min_y: int,
                       lookup_image: Optional[np.ndarray] = None) -> Tuple[int, int, np.ndarray]:
    assert not remove_by_cluster

    number_of_box_points: int = 0
    number_of_box_border_points: int = 0

    clusters: np.array = np.zeros(object_mask.shape, np.uint16)
    assert object_mask.shape == clusters.shape
    assert len(xyz_image.shape) == 3
    assert xyz_image.shape[:2] == confidence_image.shape

    number_of_elements_in_cluster: Dict[np.uint16, int] = dict()
    combined_clusters_ptr: Dict[np.uint16, np.uint16] = dict()

    cluster_value_type = np.int64
    combined_clusters_value: List[Set[cluster_value_type]] = []

    next_free_cluster_value: np.uint16 = np.uint16(1)

    assert lookup_image is None or lookup_image.shape == object_mask.shape

    for i_y in range(object_mask.shape[0]):
        for i_x in range(object_mask.shape[1]):

            if object_mask[i_y, i_x]:
                coordinate_xyz_image = i_y + min_y, i_x + min_x
                confidence = confidence_image[coordinate_xyz_image[0], coordinate_xyz_image[1]]
                xyz = xyz_image[coordinate_xyz_image[0], coordinate_xyz_image[1]]
                assert xyz.shape == (3,)

                if confidence > 0 and (not z_threshold or (abs(xyz[-1] - median_z) < z_threshold)):

                    if remove_by_cluster:
                        cluster: np.uint16 = next_free_cluster_value
                        for yy in range(max(i_y - 2, 0), i_y + 1):
                            for xx in range(max(i_x - 2, 0), i_x + 1):
                                diff = xyz[:2] - xyz_image[yy + min_y, xx + min_x, :2]
                                assert diff.shape == (2,)

                                if (diff >= 0).all() and (diff <= 10).all():
                                    other_cluster: np.uint16 = clusters[yy, xx]
                                    if other_cluster:
                                        if cluster == next_free_cluster_value or cluster == other_cluster:
                                            cluster = np.uint16(other_cluster)
                                        else:
                                            if cluster not in combined_clusters_ptr or other_cluster not \
                                                    in combined_clusters_value[combined_clusters_ptr[cluster]]:
                                                if cluster not in combined_clusters_ptr:
                                                    combined_clusters_ptr[cluster] = cluster_value_type(
                                                        len(combined_clusters_value))
                                                    combined_clusters_value.append({cluster_value_type(other_cluster)})
                                                else:
                                                    combined_clusters_value[combined_clusters_ptr[cluster]].add(
                                                        cluster_value_type(other_cluster))

                                                if other_cluster not in combined_clusters_ptr:
                                                    combined_clusters_ptr[other_cluster] = cluster_value_type(
                                                        len(combined_clusters_value))
                                                    combined_clusters_value.append({cluster_value_type(cluster)})

                                                else:
                                                    combined_clusters_value[combined_clusters_ptr[other_cluster]].add(
                                                        cluster_value_type(cluster))

                        if cluster == next_free_cluster_value:
                            number_of_elements_in_cluster[cluster] = 0
                            next_free_cluster_value += 1
                        number_of_elements_in_cluster[cluster] += 1

                        if cluster not in combined_clusters_ptr:
                            combined_clusters_ptr[cluster] = cluster_value_type(len(combined_clusters_value))
                            combined_clusters_value.append({cluster_value_type(cluster)})
                        clusters[i_y, i_x] = cluster

                    else:
                        cluster = np.uint16(0)

                    box_points[number_of_box_points, :-1] = xyz.astype(box_points.dtype)
                    box_points[number_of_box_points, -1] = box_points.dtype.type(cluster)
                    number_of_box_points += 1

                    object_mask[i_y, i_x] = 1
                else:
                    object_mask[i_y, i_x] = 0

    for i_y in range(object_mask.shape[0]):
        border_point_y = i_y == 0 or i_y + 1 == object_mask.shape[0]
        for i_x in range(object_mask.shape[1]):
            border_point = border_point_y or (i_x == 0 or i_x + 1 == object_mask.shape[1])

            if object_mask[i_y, i_x]:
                coordinate_xyz_image = i_y + min_y, i_x + min_x
                if not border_point:
                    bps = 0
                    bpso = 0
                    for i in [-1, 1]:
                        assert 0 <= i_y + i < object_mask.shape[0]
                        assert 0 <= i_x + i < object_mask.shape[1]
                        bps += object_mask[i_y, i_x + i]
                        bps += object_mask[i_y + i, i_x]
                        bpso += object_mask[i_y + i, i_x + i]
                        bpso += object_mask[i_y - i, i_x + i]
                    border_point = bps <= 2 and (bps + bpso) <= 4

                if border_point:
                    xyz = xyz_image[coordinate_xyz_image[0], coordinate_xyz_image[1]]
                    box_border_points[number_of_box_border_points, :-1] = xyz.astype(box_points.dtype)
                    box_border_points[number_of_box_border_points, -1] = 1
                    number_of_box_border_points += 1

                if lookup_image is not None:
                    lookup_image[i_y, i_x] = border_point + 1

        """
        if remove_by_cluster and number_of_box_border_points >= 4:
            cleaned_clusters: List[Tuple[int, Set[int]]] = []
            while len(combined_clusters_ptr):
                cluster_parent, unprocessed_children = next(itertools.islice(combined_clusters.items(), 0, 1))
                count = number_of_elements_in_cluster[cluster_parent]

                processed_children = {cluster_parent}
                while len(unprocessed_children):
                    current_child = unprocessed_children.pop()

                    if current_child not in processed_children:
                        count += number_of_elements_in_cluster[current_child]
                        processed_children.add(current_child)
                        unprocessed_children |= combined_clusters.pop(current_child)

                del combined_clusters[cluster_parent]
                cleaned_clusters.append((count, processed_children))

            cluster = sorted(cleaned_clusters, key=lambda a: a[0])[-1] if len(cleaned_clusters) > 0 else None

            last_free_point = 0
            for current_point in range(number_of_box_border_points):
                if box_border_points[current_point, -1] in cluster[1]:
                    if last_free_point != current_point:
                        box_border_points[last_free_point] = box_border_points[current_point]
                    last_free_point += 1
            number_of_box_border_points = last_free_point

            last_free_point = 0
            for current_point in range(number_of_box_points):
                if box_points[current_point, -1] in cluster[1]:
                    if last_free_point != current_point:
                        box_points[last_free_point] = box_points[current_point]
                    last_free_point += 1
            number_of_box_points = last_free_point
            XXX
        """
    return number_of_box_points, number_of_box_border_points, clusters


def extract_3d(detection: Dict, plot_identifier, plot_show, default_object_depth,
               object_depth_comparator_pc: Optional[np.ndarray], use_cached: bool = True):
    # Recover information from the detection
    times = [(time.time(), '')]
    localization: Optional[Dict] = detection.get('localization')
    if localization is None:
        return

    object_mask: np.ndarray = localization.get('object_mask')
    if object_depth_comparator_pc is not None and object_mask is None:
        return None

    localization_3d = localization.get('3d')
    if localization_3d is None:
        return

    # Already computed, use cached value
    if localization.get('pose3d') is not None and use_cached:
        return

    box_points: np.ndarray = localization_3d['box_points']
    box_border_points: np.ndarray = localization_3d['box_border_points']

    data = box_points[..., :3]
    data_hull_candidates = box_border_points[..., :3]

    number_of_box_points: int = data.shape[0]
    assert data.shape == (number_of_box_points, 3)

    number_of_box_border_points: int = data_hull_candidates.shape[0]
    assert data_hull_candidates.shape == (number_of_box_border_points, 3)

    if number_of_box_border_points < 4:
        return None

    # Visualize box points and box_border points
    if plot_identifier is not None or plot_show:
        # Visualize the points that are contained int he box and the hull points.
        import matplotlib.pyplot as plt
        plt.figure(figsize=(3, 1.5))
        plt.subplots_adjust(left=0, bottom=0, right=1, top=0.95, wspace=0.01, hspace=0.01)

        ax = plt.subplot(1, 2, 1)
        ax.scatter(*data_hull_candidates[..., :2].T, color='blue', s=50, alpha=0.5)
        ax.scatter(*data[..., :2].T, color='green', s=5)

    # Fit a plane.
    c, normal, mat = fit_plane(data.copy())
    times.append((time.time(), 'fit plane'))

    # Compute the convex hull in the plane.
    convex_hull_2d = convex_hull(data_hull_candidates[:, :2].copy())
    number_of_convex_hull_points: int = convex_hull_2d.shape[0]
    assert convex_hull_2d.shape == (number_of_convex_hull_points, 2), convex_hull_2d.shape
    times.append((time.time(), 'convex hull'))

    # Compute the minimum enclosing rectangle (size (2d) and angle).
    angle: float
    center, size, angle = get_minimum_enclosing_rectangle(convex_hull_2d.copy(), data,
                                                          plot_identifier=plot_identifier, plot_show=plot_show)
    times.append((time.time(), 'min rectangle hull'))
    assert size.shape == (2,)
    assert center.shape == (2,)

    # Computer center (3d),
    # center = np.mean(data, axis=0)
    center = np.append(center, 0)
    assert center.shape == (3,)
    center = project_onto_plane(normal, c, vectors=center[None])

    pose3d = np.empty((2, 3), dtype=data.dtype)
    pose3d[0] = center

    # roll = np.atan(b/c)
    pose3d[1] = [0, 0, angle]  # < euler angles

    size3d = np.zeros(3, dtype=data.dtype)
    size3d[:2] = size

    size3d[2] = default_object_depth
    if object_depth_comparator_pc is not None:
        assert object_mask is not None
        min_x, min_y = localization['min_x'], localization['min_y']
        max_x, max_y = localization['max_x'], localization['max_y']

        relevant_old_z = object_depth_comparator_pc[min_y:max_y, min_x:max_x, -1]
        relevant_old_z[relevant_old_z == 0] = 2250 - 40  # XXX: replace 2250
        old_z = np.median(relevant_old_z)
        new_z = np.median(data[..., -1])
        size3d[2] = old_z - new_z

    pose3d = pose3d.tolist()
    size3d = size3d.tolist()
    detection['localization']['pose3d'] = detection['pose3d'] = pose3d
    detection['localization']['size3d'] = detection['size3d'] = size3d

    logger.debug('Package Matching: Detail: extract_3d times'
                 + ', '.join([f'{v2}: {int((t2 - t1) * 1000) / 1000}'
                              for (t1, _), (t2, v2) in zip(times[:-1], times[1:])])
                 + f' total: {times[-1][0] - times[0][0]}')


def convex_hull(points: np.ndarray) -> np.ndarray:
    n_points: int = points.shape[0]
    dimensionality: int = points.shape[1]
    assert points.shape == (n_points, dimensionality), points.shape
    assert n_points >= 3
    assert dimensionality == 2
    points = points[np.lexsort(points.T)]

    return _convex_hull(points)


# XXX: maybe use int32 to be more certaint hat everything is ok
# njit!
@numba.jit(numba.float32[:, :](numba.float32[:, :]), nopython=True, parallel=False)
def _convex_hull(points: np.ndarray) -> np.ndarray:
    """
    Compute the convex hull of the points using the monotone chain algorithm
    :param points:
    sorted!!!
    :return:
    """
    n_points: int = points.shape[0]
    dimensionality: int = points.shape[1]
    # assert points.shape == (n_points, dimensionality), points.shape
    # assert n_points >= 3
    # assert dimensionality == 2

    upper_hull = np.empty_like(points)
    lower_hull = np.empty((n_points * 2, dimensionality), dtype=points.dtype)
    i_uh: int = 0
    i_lh: int = 0

    for p in points:
        while i_lh >= 2:
            origin, p1, p2 = lower_hull[i_lh - 2], lower_hull[i_lh - 1], p
            if cross2d(p1 - origin, p2 - origin) > 0:
                break
            i_lh -= 1
        lower_hull[i_lh] = p
        i_lh += 1

    for p in points[::-1]:
        while i_uh >= 2:
            origin, p1, p2 = upper_hull[i_uh - 2], upper_hull[i_uh - 1], p
            if cross2d(p1 - origin, p2 - origin) > 0:
                break
            i_uh -= 1

        upper_hull[i_uh] = p
        i_uh += 1

    # Remove the last point of each list because it is the same as the first point of the other list
    # i_uh -= 1
    i_lh -= 1

    convex_hull_arr = lower_hull[:i_uh + i_lh]
    convex_hull_arr[i_lh:] = upper_hull[:i_uh]
    return convex_hull_arr


def reduce_convex_hull(convex_hull_complete: np.ndarray, threshold: float = 10.0):
    convex_hull_reduced = convex_hull_complete.copy()
    last_point_i = 0
    num_accepted: int = 1
    for current_i, point in enumerate(convex_hull[1:], 1):
        if current_i <= last_point_i + 1:
            continue

        distances = distance(convex_hull_complete[last_point_i], convex_hull_complete[current_i],
                             convex_hull_complete[last_point_i + 1:current_i])

        if (distances > threshold).any():
            # Accept the point
            last_point_i = current_i
            convex_hull_reduced[num_accepted] = point
            num_accepted += 1

    return convex_hull_reduced[:num_accepted]


def get_minimum_enclosing_rectangle(convex_hull: np.ndarray, data: Optional[np.ndarray],
                                    plot_identifier: Optional[str] = None, plot_show: bool = False) \
        -> Tuple[Optional[np.ndarray], Optional[np.ndarray], Optional[float]]:
    number_of_points_in_convex_hull: int = convex_hull.shape[0]
    assert convex_hull.shape == (number_of_points_in_convex_hull, 2)

    # Remove points from the convex hull that don't matter too much.
    # convex_hull = reduce_convex_hull(convex_hull)
    # number_of_points_in_convex_hull: int = convex_hull.shape[0]
    # assert convex_hull.shape == (number_of_points_in_convex_hull, 2)

    vectors = np.diff(convex_hull, axis=0)
    number_of_vectors: int = vectors.shape[0]
    assert number_of_vectors == number_of_points_in_convex_hull - 1
    assert vectors.shape == (number_of_vectors, 2), vectors.shape

    # The angles of all the vectors in the convex hull.
    angles = -get_angles_from_points(vectors)
    assert angles.shape == (number_of_vectors,)

    best_size: Optional[np.ndarray] = None
    smallest_area: Optional[float] = None
    best_angle: Optional[float] = None
    best_center: Optional[np.ndarray] = None

    angles = sorted(angles)

    # tn = time.time()
    # print(f'\t {tn - tl} beginning stuff with {len(angles)} took')
    # tl = tn

    # Remove points with similar orientations
    last_angle = None
    for angle_i, angle in enumerate(angles):
        if last_angle is not None and angle - last_angle < 0.1:
            continue

        last_angle = angle

        rotated_convex_hull = rot_points_multi_rads(convex_hull.copy()[None], angle[None])
        assert rotated_convex_hull.shape == (1, convex_hull.shape[0], 2), rotated_convex_hull.shape
        # tn = time.time()
        # print(f'\t {tn - tl} rotation took')
        # tl = tn

        rotated_convex_hull = rotated_convex_hull[0]
        assert rotated_convex_hull.shape == (convex_hull.shape[0], 2), rotated_convex_hull.shape

        min_xy = np.min(rotated_convex_hull, axis=0)
        max_xy = np.max(rotated_convex_hull, axis=0)
        size = max_xy - min_xy
        center = (min_xy + max_xy) / 2.0

        assert min_xy.shape == (2,), min_xy.shape
        assert max_xy.shape == (2,), max_xy.shape
        assert size.shape == (2,), size.shape
        assert center.shape == (2,), center.shape
        # BUG: area = np.linalg.norm(size)
        area = size[0] * size[1]

        def show():
            # nc = center.copy()
            # nc = rot_points_multi_rads(nc[None, None], -angle[None])

            rotated_data = rot_points_multi_rads(data[:, :2].copy()[None], angle[None])
            assert rotated_data.shape == (1, data.shape[0], 2), rotated_data.shape
            rotated_data = rotated_data[0]
            assert rotated_data.shape == (data.shape[0], 2), rotated_data.shape

            import matplotlib.pyplot as plt
            plt.figure()
            ax = plt.subplot(2, 2, 1)
            ax.scatter(*data[..., :2].T, color='green', s=5)
            ax.scatter(*convex_hull.T, color='blue', s=50, alpha=0.5)
            ax.set_aspect('equal', adjustable='box')

            ax = plt.subplot(2, 2, 2)
            ax.scatter(*rotated_convex_hull.T, color='blue', s=50, alpha=0.5)
            ax.scatter(*rotated_data[..., :2].T, color='green', s=5)
            ax.set_aspect('equal', adjustable='box')

            ax = plt.subplot(2, 2, 3)
            ax.scatter(*center.T, color='red', s=500)
            ax.scatter(*rotated_convex_hull.T, color='blue', s=50, alpha=0.5)
            ax.scatter(*rotated_data[..., :2].T, color='green', s=5)
            ax.set_aspect('equal', adjustable='box')

            print(area, size)
            plt.show()

        # tn = time.time()
        # print(f'\t {tn - tl} simple computation')
        # tl = tn

        """
        mimi = rotated_convex_hull - min_xy
        assert (mimi >= -1e-4).all()
        mama = max_xy - rotated_convex_hull
        assert (mama >= -1e-4).all()

        assert ((rotated_convex_hull - (center - size / 2)) >= -1e-4 ).all()
        cps2 = center + size / 2 - rotated_convex_hull
        assert (cps2 >= -1e-4).all()

        """
        if best_size is None or area < smallest_area:
            best_size = size
            smallest_area = area

            best_angle = angle
            best_center = center.copy()
            best_center = rot_points_multi_rads(best_center[None, None], -best_angle[None])

            # tn = time.time()
            # print(f'\t {tn - tl} save best')
            # tl = tn

    if plot_identifier is not None or plot_show:
        import matplotlib.pyplot as plt
        ax = plt.subplot(1, 2, 2)
        ax.scatter(*best_center.T, color='red', s=500)
        ax.scatter(*convex_hull.T, color='blue', s=50, alpha=0.5)
        ax.scatter(*data[..., :2].T, color='green', s=5)

        # """
        distance_difference: np.ndarray = np.diag(best_size)[None] / 2
        assert distance_difference.shape == (1, 2, 2)
        __tmp = rot_points_multi_rads(points=distance_difference, rotation_rads=np.array(-best_angle)[None])
        assert np.all(
            __tmp == distance_difference), 'Rotations are supposed to happen inplace, __tmp is therefore just ' \
                                           'a view to distance_difference.'
        assert distance_difference.shape == (1, 2, 2)

        to_top = distance_difference[0, 0]
        to_right = distance_difference[0, 1]
        borders = best_center + np.array(
            [to_top - to_right, to_top + to_right, -to_top + to_right, -to_top - to_right, to_top - to_right])
        ax.plot(*borders.T, color='red', linestyle='--')

        ax.set_aspect('equal')
        plt.suptitle(f'{"x".join(map(str, map(int, best_size)))}')

        if plot_identifier:
            plt.savefig(f'/home/dataspree/.dataspree/inference/output/{plot_identifier}_3d_localization.png')
        if plot_show:
            plt.show()

    best_center = best_center[0, 0]

    assert best_center is None or best_center.shape == (2,), best_center.shape
    assert best_size is None or best_size.shape == (2,), best_size.shape

    return best_center, best_size, -best_angle


def distance(last_point: np.ndarray, next_point: np.ndarray, current_point: np.ndarray):
    assert last_point.shape == (2,)
    assert next_point.shape == (2,)
    number_of_points: int = current_point.shape[0]
    assert current_point.shape == (number_of_points, 2)

    denuminator = np.linalg.norm(next_point - current_point, axis=1)
    assert denuminator.shape == (number_of_points,)

    nominator = np.cross(next_point[None] - current_point, current_point - last_point)
    assert nominator.shape == (number_of_points,)

    return np.abs(nominator) / denuminator


def dataspree_inference_boxes_to_vectors(detections: List[Dict[str, Any]], image_width: int, image_height: int,
                                         threshold: float,
                                         specific_threshold: Optional[Dict] = None) -> Tuple[np.ndarray, np.ndarray]:
    """
    Return the coordinates in the image of the boxes.
    """
    number_of_boxes: int = len(detections)

    if specific_threshold is None:
        specific_threshold = dict()

    ai_platform_box_indices = np.empty((number_of_boxes,), dtype=np.int32)
    ai_platform_boxes = np.empty((number_of_boxes, 5), dtype=np.float32)
    index: int = 0
    for detection_index, detection in enumerate(detections):
        if detection['confidence'] >= specific_threshold.get(detection['label'], threshold):
            ai_platform_boxes[index] = detection['x'] * image_width, detection['y'] * image_height, \
                                       detection['width'] * image_width, detection['height'] * image_height, \
                                       detection['orientation']
            ai_platform_box_indices[index] = detection_index
            index += 1

    box_representations = ai_platform_boxes_to_vectors(ai_platform_boxes[:index])
    assert box_representations.shape == (index, 5, 2)

    np.clip(box_representations, [0, 0], [image_width - 1, image_height - 1], out=box_representations)

    ai_platform_box_indices = ai_platform_box_indices[:index]
    assert ai_platform_box_indices.shape == (index,)

    for original_box_index, box_representation in zip(ai_platform_box_indices, box_representations):
        assert box_representation.shape == (5, 2)
        detection = detections[original_box_index]

        localization_dict = detection.get('localization')
        if localization_dict is None:
            localization_dict = detection['localization'] = dict()

        # box_representation:  bottom_left, top_left, top_right, min_values, max_values
        # rotated_box:  bottom_left, top_left, top_right
        localization_dict['rotated_box'] = box_representation[:3]

        # enclosing_rectangle:  min_values, max_values
        enclosing_rectangle = localization_dict['enclosing_rectangle'] = np.round(box_representation[3:]).astype(int)
        min_x, min_y = localization_dict['min_x'], localization_dict['min_y'] = enclosing_rectangle[0]
        max_x, max_y = localization_dict['max_x'], localization_dict['max_y'] = enclosing_rectangle[1]
        c_x, c_y = localization_dict['c_x'], localization_dict['c_y'] \
            = int(round((min_x + max_x) / 2.0)), int(round((min_y + max_y) / 2.0))

        assert 0 <= min_x <= c_x <= max_x <= image_width, (0, min_x, c_x, max_x, image_width)
        assert 0 <= min_y <= c_y <= max_y <= image_height, (0, min_y, c_y, max_y, image_height)

    return box_representations, ai_platform_box_indices


def ai_platform_boxes_to_vectors(ai_platform_boxes: np.ndarray) -> np.ndarray:
    """
    :param ai_platform_boxes:      boxes in the Dataspree AI Platform's format (shape number of boxes x 5)
    """
    number_of_boxes: int = ai_platform_boxes.shape[0]
    assert ai_platform_boxes.shape == (number_of_boxes, 5)

    centers, distances = ai_platform_boxes[:, :2], ai_platform_boxes[:, 2:4] / 2
    rotations = (ai_platform_boxes[:, 4] * (-1) + np.pi) % (2 * np.pi) - np.pi  # + np.pi
    assert centers.shape == (number_of_boxes, 2)
    assert distances.shape == (number_of_boxes, 2)
    assert rotations.shape == (number_of_boxes,)

    distance_difference: np.ndarray = np.array([distances * np.array([0, 1]), distances * np.array([1, 0])]) \
        .reshape((2, number_of_boxes, 2)).swapaxes(0, 1)
    assert distance_difference.shape == (number_of_boxes, 2, 2)
    __tmp = rot_points_multi_rads(points=distance_difference, rotation_rads=rotations)

    if is_application_debug() and __debug__:
        assert np.all(__tmp == distance_difference), \
            'Rotations are supposed to happen inplace, __tmp is therefore just a view to distance_difference.'

    assert distance_difference.shape == (number_of_boxes, 2, 2)

    to_bottom = distance_difference[:, 0]
    to_right = distance_difference[:, 1]

    top = centers - to_bottom
    right = centers + to_right
    assert top.shape == (number_of_boxes, 2) and right.shape == (number_of_boxes, 2) and \
           centers.shape == (number_of_boxes, 2) and rotations.shape == (number_of_boxes,)

    top_left = centers - to_bottom - to_right
    top_right = centers - to_bottom + to_right
    bottom_right = centers + to_bottom + to_right
    bottom_left = centers + to_bottom - to_right

    box_corners = np.stack((top_left, top_right, bottom_right, bottom_left), axis=1)

    # import matplotlib.pyplot as plt
    # plt.figure()
    # plt.plot(*box_corners.T, color = 'red')
    # plt.scatter(*np.array([image_width, image_height])[None].T)
    # plt.show()

    assert box_corners.shape == (number_of_boxes, 4, 2), box_corners.shape

    min_values = np.min(box_corners, axis=1)
    max_values = np.max(box_corners, axis=1) + 1
    assert min_values.shape == (number_of_boxes, 2), min_values.shape
    assert max_values.shape == (number_of_boxes, 2), max_values.shape

    box_representation = np.stack((bottom_left, top_left, top_right, min_values, max_values), axis=1)
    box_representation.shape = (number_of_boxes, 5, 2)
    return box_representation


def rot_points_multi_rads(points: np.ndarray, rotation_rads: np.ndarray) -> np.ndarray:
    """
    Rotate the box points inplace (!) and return a reference to the first argument.

    :param points:         Points to be rotated (Shape: number_of_rotations, number_of_points_per_rotation, 2)
                           Attention: The rotation is performed inplace; hence the content of the "points" is
                           overwritten.
                           If you would like to keep the unrotated points, call this function on a copy of the points
                           numpy array.

    :param rotation_rads:  Angles along which points are rotated. (Shape: number_of_rotations)

    :return:               Rotated points (Shape: number_of_rotations, number_of_points_per_rotation, 2) to offer
                           the user the ability to call this function with a "rvalue" or a copy.

    @dev: Implementation taken from Dataspree AI Platform Training.
          TODO: include all utilities into the processing lib or the platform sdk
    """
    number_of_rotations: int = points.shape[0]
    number_of_points_to_rotate_per_rotation: int = points.shape[1]
    state_dimensionality: int = points.shape[2]
    assert points.shape == (number_of_rotations, number_of_points_to_rotate_per_rotation, state_dimensionality)
    assert rotation_rads.shape == (number_of_rotations,)

    cr: np.ndarray = np.cos(rotation_rads)
    sr: np.ndarray = np.sin(rotation_rads)
    rotation_matrix = np.array([[cr, -sr], [sr, cr]]).reshape((2, 2, number_of_rotations))
    assert rotation_matrix.shape == (state_dimensionality, state_dimensionality, number_of_rotations)

    is_debug: bool = __debug__ and is_application_debug()
    if is_debug:
        __points = points.copy()

    np.einsum('rcR,Rxc->Rxr', rotation_matrix, points, out=points)
    assert points.shape == (number_of_rotations, number_of_points_to_rotate_per_rotation, state_dimensionality)

    if is_debug:
        for i in range(len(__points)):
            # noinspection PyUnreachableCode
            __alt = rotation_matrix[:, :, i].dot(__points[i].swapaxes(0, 1)).swapaxes(0, 1)
            _difference = np.abs(__alt - points[i])
            assert np.all(_difference < 1e-2)

    return points


def coordinate_transformation(spanning_points: np.ndarray) \
        -> Tuple[Callable[[Union[np.ndarray, List[np.ndarray]]], np.ndarray],
                 Callable[[Union[np.ndarray, List[np.ndarray]]], np.ndarray]]:
    """
    Compute functions for conducting a coordinate transformation from the canonical euclidean coordinate system
    to the coordinate system spanned by the #spanning_points.

    :param spanning_points: Three points (A, B, C), each \\in \\R^2 that span a two-dimensional coordinate system
                            through the vectors BA, BC.

    :return:                Tuple: Function to conduct transformation into the new coordinate system,
                                   function to convert backwards.

                            Each function takes as input a numpy array of size 2 or (n x 2), and emits an array
                            of the input size.
    """
    number_of_spanning_points: int = spanning_points.shape[0]
    dimensionality: int = spanning_points.shape[1]
    assert spanning_points.shape == (number_of_spanning_points, dimensionality)
    assert dimensionality + 1 == number_of_spanning_points
    assert dimensionality == 2, 'The dimensionality of the points has to be two, otherwise slightly adapt and test.'

    new_origin: np.ndarray = spanning_points[1]
    assert new_origin.shape == (dimensionality,)

    base_system = (spanning_points[[0, 2]] - new_origin).T
    assert base_system.shape == (dimensionality, dimensionality)

    coordinate_transform_matrix = np.linalg.pinv(base_system)
    assert coordinate_transform_matrix.shape == (dimensionality, dimensionality)

    # TODO: Use einsum notation and omit .T's
    return (lambda a: coordinate_transform_matrix.dot((a - new_origin).T).T,
            lambda a: base_system.dot(a.T).T + new_origin)


def get_angles_from_points(points: np.ndarray) -> np.ndarray:
    """
    Return the angles from a set of two-dimensional points

    :param points:      points (shape N x 2)
    :return:            angles of points (shape N)
    """
    number_of_points: int = points.shape[0]
    point_dimensionality: int = points.shape[1]
    assert points.shape == (number_of_points, point_dimensionality)
    assert point_dimensionality == 2, \
        f'Can only get angles from two-dimensional points. Dimensionality: {point_dimensionality} '

    angles = np.arctan2(*points[:, ::-1].T)
    assert angles.shape == (number_of_points,)
    return angles


def fit_plane(xyz_image):
    mat_g = xyz_image.copy()
    mat_g[..., -1] = 1

    # np.linalg.

    try:
        mat, resid, rank, s = np.linalg.lstsq(mat_g, xyz_image[:, 2], rcond=None)
    except ValueError as e:
        raise e from None

    a, b, c = mat

    normal = (a, b, -1)
    return c, normal / np.linalg.norm(normal), mat


def project_onto_plane(normal, c, *, xx=None, yy=None, vectors=None):
    if vectors is None:
        return (-normal[0] * xx - normal[1] * yy - np.array([0.0, 0.0, c]).dot(normal)) * 1. / normal[2]
    vectors[:, 2] = (-normal[0] * vectors[:, 0] - normal[1] * vectors[:, 1] - np.array([0.0, 0.0, c]).dot(normal)) \
                    * 1. / normal[2]
    return vectors


def visualize_plane_fit(data, normal, c, convex_hull_2d, size, angle, center):
    # return
    import matplotlib.pyplot as plt
    plt.figure(figsize=(12, 5))
    ax = plt.subplot(121, projection='3d')

    # Visualize the plane all by itself (without surface rotation or boundraries)
    maxx, maxy = np.max(data[:, 0]), np.max(data[:, 1])
    minx, miny = np.min(data[:, 0]), np.min(data[:, 1])
    xx, yy = np.meshgrid([minx, maxx], [miny, maxy])
    zz = project_onto_plane(normal, c, xx=xx, yy=yy)
    ax.plot_surface(xx, yy, zz, alpha=0.2)

    # Scatter plot of the data projected onto the plane
    projected_data = data.copy()
    project_onto_plane(normal, c, vectors=projected_data)
    ax.scatter(*projected_data.T, color='green', s=5)

    # Plot the projection of the convex hull
    convex_hull_3d = np.empty((convex_hull_2d.shape[0], 3), dtype=convex_hull_2d.dtype)
    convex_hull_3d[:, :2] = convex_hull_2d
    project_onto_plane(normal, c, vectors=convex_hull_3d)
    ax.plot(*convex_hull_3d.T, color='black')

    # Plot the minimum enclosing rectangle
    assert center.shape == (3,), center.shape
    distance_difference: np.ndarray = np.diag(size)[None] / 2
    assert distance_difference.shape == (1, 2, 2)
    __tmp = rot_points_multi_rads(points=distance_difference, rotation_rads=np.array(angle)[None])

    if __debug__ and is_application_debug():
        assert np.all(__tmp == distance_difference), \
            'Rotations are supposed to happen inplace, __tmp is therefore just a view to distance_difference.'
    assert distance_difference.shape == (1, 2, 2)

    to_top = np.zeros(3, dtype=center.dtype)
    to_right = np.zeros(3, dtype=center.dtype)
    to_top[:2] = distance_difference[0, 0]
    to_right[:2] = distance_difference[0, 1]
    borders = center + np.array(
        [to_top - to_right, to_top + to_right, -to_top + to_right, -to_top - to_right, to_top - to_right])
    project_onto_plane(normal, c, vectors=borders)
    ax.plot(*borders.T, color='black', linestyle='--')

    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')

    sp2 = plt.subplot(1, 2, 2)
    sp2.scatter(*data[:, :2].T)
    sp2.plot(*convex_hull_2d.T, color='black')
    sp2.plot(*borders[:, :2].T, color='black', linestyle='--')
    sp2.scatter(*center[:2].T, color='red', s=100)

    plt.show()
    ...


def center_size_rotation_to_box_points(center, size, rotation):
    distance_difference: np.ndarray = np.diag(size)[None] / 2
    assert distance_difference.shape == (1, 2, 2)
    __tmp = rot_points_multi_rads(points=distance_difference, rotation_rads=np.array(-rotation)[None])
    assert np.all(
        __tmp == distance_difference), 'Rotations are supposed to happen inplace, __tmp is therefore just ' \
                                       'a view to distance_difference.'
    assert distance_difference.shape == (1, 2, 2)

    to_top = distance_difference[0, 0]
    to_right = distance_difference[0, 1]
    #top_left, top_right, bottom_right, bottom_left
    return center + np.array(
        [to_top - to_right, to_top + to_right, -to_top + to_right, -to_top - to_right])

def get_box_direction_vectors(size, rotation) -> Tuple[np.ndarray, np.ndarray]:
    distance_difference: np.ndarray = np.diag(size)[None] / 2
    assert distance_difference.shape == (1, 2, 2)
    __tmp = rot_points_multi_rads(points=distance_difference, rotation_rads=np.array(-rotation)[None])
    assert np.all(
        __tmp == distance_difference), 'Rotations are supposed to happen inplace, __tmp is therefore just ' \
                                       'a view to distance_difference.'
    assert distance_difference.shape == (1, 2, 2)

    to_top = distance_difference[0, 0]
    to_right = distance_difference[0, 1]
    return to_top, to_right
