from typing import Optional, Dict, List, Tuple, Union

import numpy as np
import matplotlib.pyplot as plt
import scipy.optimize

from dataspree.inference.plugins.tracking.mbm import MultibernoulliBirthModel, \
    GaussianPredictionParametrization, UpdateModel


class PoissonPointProcessBirthModel(MultibernoulliBirthModel):
    ...


def expand_array(arr: np.ndarray, new_length: int, *, refcheck: bool = False):
    """
    Expand a numpy array and ensure that the requested memory is actually allocated

    :param arr:
    :param new_length:
    :param refcheck:
    :return:
    """
    old_shape = arr.shape
    arr.resize((new_length, *old_shape[1:]), refcheck=refcheck)
    # arr[old_shape[0]:].fill(0)


def allocate_array(shape: Union[Tuple, List[int]], **kwargs) -> np.ndarray:
    a = np.empty(0, **kwargs)
    a.resize(shape)
    a.fill(0)
    return a


class PMBM:
    """
    PMBM advantage over solely Multi Bernoulli conjugate priors:
    - in MBM we have to add birth parameters to each a fixed number of birth components to each Multibernoulli in
      the mixture
    - in PMBM, by modelling

    """

    def __init__(self,
                 predict: GaussianPredictionParametrization,
                 update: UpdateModel,
                 max_number_of_hypothesis,
                 probability_of_detection: float = 0):
        self.predict: GaussianPredictionParametrization = predict
        self.update: UpdateModel = update

        state_dimensionality: int = predict.birth.get_state_dimensionality()

        # Reserve memory for the undetected components
        number_of_undetected_components_start = 10 ** 3
        self.number_of_undetected_objects: int = 0
        self.undetected_labels: np.ndarray = allocate_array((number_of_undetected_components_start, 2))
        self.undetected_weights: np.ndarray = allocate_array((number_of_undetected_components_start,))
        self.undetected_means: np.ndarray = allocate_array((number_of_undetected_components_start,
                                                            state_dimensionality))
        self.undetected_covariance_matrices: np.ndarray = allocate_array((number_of_undetected_components_start,
                                                                          state_dimensionality, state_dimensionality))

        # Reserve some memory for the detected components
        self.number_of_hypothesis: int = 0
        self.log_weights: np.ndarray = allocate_array((max_number_of_hypothesis ** 2), dtype=np.float)
        self.number_of_detected_objects: np.ndarray = allocate_array((max_number_of_hypothesis ** 2), dtype=np.int)
        self.labels: List[np.ndarray] = []
        self.probabilities_of_existence: List[np.ndarray] = []
        self.means: List[np.ndarray] = []
        self.covariance_matrices: List[np.ndarray] = []

        self.time_count: int = 0

        ...

    def __predict(self, dt: float) -> None:

        #
        # Undetected objects
        #
        number_of_undetected_objects: int = self.number_of_undetected_objects

        # Predict the existing components
        self.predict.transition.predict(
            probabilities_of_existence=self.undetected_weights[:number_of_undetected_objects],
            means=self.undetected_means[:number_of_undetected_objects],
            covariances=self.undetected_covariance_matrices[:number_of_undetected_objects], dt=dt)

        # Allocate memory if necessary
        number_of_birth_components: int = self.predict.birth.get_number_of_birth_components()
        new_number_of_undetected_objects: int = number_of_undetected_objects + number_of_birth_components
        if self.undetected_labels.shape[0] <= new_number_of_undetected_objects:
            number_of_elements = self.undetected_labels.shape[0]
            assert self.undetected_labels.shape == (number_of_elements, 2)
            assert self.undetected_weights.shape == (number_of_elements,)

            state_dimensionality = self.undetected_means.shape[1]
            assert self.undetected_means.shape == (number_of_elements, state_dimensionality)
            assert self.undetected_covariance_matrices.shape == (number_of_elements, state_dimensionality,
                                                                 state_dimensionality)

            new_number_of_elements: int = number_of_elements * 2
            expand_array(self.undetected_labels, new_number_of_elements)
            expand_array(self.undetected_weights, new_number_of_elements)
            expand_array(self.undetected_means, new_number_of_elements)
            expand_array(self.undetected_covariance_matrices, new_number_of_elements)

            assert self.undetected_labels.shape == (new_number_of_elements, 2)
            assert self.undetected_weights.shape == (new_number_of_elements,)
            assert self.undetected_means.shape == (new_number_of_elements, state_dimensionality)
            assert self.undetected_covariance_matrices.shape == (new_number_of_elements, state_dimensionality,
                                                                 state_dimensionality)

        # Add birth components and increase the number 0
        self.predict.birth.predict_put(
            time_count=self.time_count, labels=self.undetected_labels[number_of_undetected_objects:],
            weights=self.undetected_weights[number_of_undetected_objects:],
            means=self.undetected_means[number_of_undetected_objects:],
            covariance_matrices=self.undetected_covariance_matrices[number_of_undetected_objects:])
        self.number_of_undetected_objects = new_number_of_undetected_objects

        #
        # Detected objects
        #
        for h in self.active_hypothesis_ids():
            number_of_detected_objects: int = self.number_of_detected_objects[h]
            self.predict.transition.predict(
                probabilities_of_existence=self.probabilities_of_existence[h][:number_of_detected_objects],
                means=self.means[h][:number_of_detected_objects],
                covariances=self.covariance_matrices[h][:number_of_detected_objects], dt=dt)

    def __update(self):

        # log_weights remain unchanged

        nw, nm, nc = self.predict.birth.predict()
        ow, om, oc = self.predict.transition.predict(self.undetected_weights, self.undetected_means,
                                                     self.undetected_cov)
        self.undetected_weights = np.vstack(nw, ow)
        self.undetected_means = np.vstack(nm, om)
        self.undetected_cov = np.vstack(nc, oc)

        self.mb_probability_existance, self.mb_mu, self.mb_cov = \
            self.predict.transition.predict(self.mb_probability_existance, self.mb_mu, self.mb_cov)

        self.inv = None

    def __get_cost_matrix(self, cost_matrix):
        ...

    def __call__(self, x: np.ndarray):
        # XXX
        if self.inv is None:
            self.inv = np.linalg.inv(self.posterior_cov)
            self.divisor = np.sqrt(2 * np.pi ** self.posterior_mu.shape[1]() * np.linalg.det(self.posterior_cov))

        return MultibernoulliBirthModel.evaluate(x, self.w, self.mu, self.inv, self.divisor)

    def show(self):
        ...
