import numpy as np

from typing import Any, Dict, List, Tuple, Optional

from albumentations import DualTransform


class ChannelAssembler(DualTransform):

    def __init__(self, channels: List[Dict[str, Any]], always_apply: bool, p: float) -> None:
        super().__init__(always_apply=always_apply, p=p)
        self.channels = channels

    def apply(self, img: np.ndarray, **_params) -> np.ndarray:
        available_channel_names: Optional[Tuple] = img.dtype.names
        is_structured_array: bool = available_channel_names is not None
        if not is_structured_array:
            # Nothing to do here.
            return img

        # target in channel last layout
        output_array = np.empty((*img.shape, len(self.channels)), dtype=np.float32)

        # Scale to 0-1 and assign to output array
        for channel_id, channel in enumerate(self.channels):
            channel_configuration = channel['channel']
            channel_name: str = channel_configuration['channel_type']['selection']
            if channel_name == 'custom':
                channel_name = channel_configuration['channel_type']['name']

            # Collect channel from image or compose it from a set of available options.
            image_channel: np.ndarray
            if channel_name in available_channel_names:
                image_channel = img[channel_name]

            elif channel_name == 'gray' or channel_name == 'a':
                image_channel = ((img['red'].astype(np.float32) + img['green'].astype(np.float32)
                                  + img['blue'].astype(np.float32)) / 3).astype(img['red'].dtype)

            else:
                raise ValueError(f'Channel {channel_name} not available. Available channels: '
                                 f'{", ".join(available_channel_names)}.')

            is_integer_dtype: bool = np.issubdtype(image_channel.dtype, np.integer)
            is_signed_integer_dtype: bool = np.issubdtype(image_channel.dtype, np.signedinteger)
            is_unsigned_integer: bool = is_integer_dtype and not is_signed_integer_dtype

            channel_operations = channel_configuration.get('enable_operations', dict()).get('operations', dict())
            enabled_operation = channel_operations.get('selection')

            min_value: float
            max_value: float
            if enabled_operation == 'absolute_cap':
                min_value: float = channel_operations['min']
                max_value: float = channel_operations['max']

            elif enabled_operation == 'relative_cap' or (not enabled_operation and is_unsigned_integer):
                if not is_unsigned_integer:
                    raise ValueError(
                        '"Relative Cap" can only be used on singed integer channels, not on channels of '
                        f'dtype {image_channel.dtype}. '
                        f'Processing channel {channel_name} failed for item.')

                max_per_datatype = np.iinfo(image_channel.dtype).max
                min_value: float = channel_operations.get('min', 0) * max_per_datatype
                max_value: float = channel_operations.get('max', 1) * max_per_datatype

            elif enabled_operation == 'normalize' or (not enabled_operation and not is_unsigned_integer):
                min_value: float = np.min(image_channel)
                max_value: float = np.max(image_channel)

            else:
                raise ValueError(f'Unknown operation "{enabled_operation}" for channel {channel_name}.')

            difference = max_value - min_value if max_value > min_value else 1

            target_dtype = np.float32
            output_array[..., channel_id] = ((image_channel.astype(np.float32, copy=False) - min_value) \
                                             / difference * 255).astype(target_dtype, copy=False)

        # (0 <= output_array).all() and (output_array <= 255).all()
        return output_array.round().astype(np.uint8)

    def apply_to_bbox(self, bbox, **_params):
        return bbox

    def apply_to_keypoint(self, keypoint, **_params):
        return keypoint

    def get_transform_init_args_names(self) -> Tuple[str]:
        return ('channels',)

    @classmethod
    def get_class_fullname(cls):
        return cls.__name__
